from math import floor
from itertools import chain
from collections import namedtuple
from reportlab.pdfbase.pdfmetrics import stringWidth, getFont, getAscentDescent

from bericht.node import Block
from .style import TextAlign

__all__ = (
    'Paragraph', 'Word', 'Break',
)


class StyledPart:
    __slots__ = ('parent', 'style', 'part')

    def __init__(self, parent, style, part):
        self.parent = parent
        self.style = style
        self.part = part

    @property
    def width(self):
        return stringWidth(self.part, self.style.font_name, self.style.font_size)


class Word:
    __slots__ = ('parts',)

    def __init__(self, paragraph, style, word):
        self.parts = [StyledPart(paragraph, style, word)]

    def add(self, paragraph, style, part):
        if self.parts[-1].style != style:
            self.parts.append(StyledPart(paragraph, style, part))
        else:
            self.parts[-1] = StyledPart(paragraph, style, self.parts[-1].part+part)
        return self

    @property
    def width(self):
        return sum(p.width for p in self.parts)

    def __str__(self):
        return ''.join(p.part for p in self.parts)


class Break:
    pass


class Line:

    __slots__ = ('words', 'width')

    def __init__(self, words=None):
        self.words = words or []
        self.width = 0

    def add(self, word):
        self.words.append(word)

    @property
    def empty(self):
        return not self.words

    def __iter__(self):
        return iter(self.words)


class Paragraph(Block):

    __slots__ = ('words',)

    def __init__(self, *args):
        super().__init__(*args)
        self.words = []

    def __str__(self):
        return ' '.join(str(word) for word in self.words)

    @property
    def min_content_width(self):
        return 0

    def wrap(self, page, available_width):
        self.width = available_width
        self.height = 0
        line = Line()
        lines = self.children = [line]
        if not self.words:
            return 0, 0

        self.css.apply(self)
        if self.parent:
            self.style = self.style.inherit(self.parent.style)

        max_width = 0
        space_width = stringWidth(' ', self.style.font_name, self.style.font_size)
        for word in self.words:

            if word is Break:
                line = Line([Break])
                lines.append(line)
                continue

            for part in word.parts:
                part.style = part.style.inherit(self.style)

            word_width = word.width
            if line.width+word_width > available_width:
                if line:
                    line = Line()
                    lines.append(line)

            line.width += word_width + space_width
            line.add(word)
            max_width = max(line.width, max_width)

        self.height = self.frame_height + len(lines) * self.style.leading
        return max_width, self.height

    def split(self, top_container, bottom_container, available_height):
        lines = floor(available_height / self.style.leading)
        if lines == len(self.children):
            self.parent = top_container
            top_container.children.append(self)
            return self, None
        else:
            top = Paragraph(self.tag, top_container, self.id, self.classes, self.css, self.position)
            top.words = list(chain(*self.children[:lines]))
            bottom = Paragraph(self.tag, bottom_container, self.id, self.classes, self.css, self.position)
            bottom.words = list(chain(*self.children[lines:]))
            return top, bottom

    def draw(self, page, x, y):
        final_x, final_y = x, y - self.height
        style = self.style
        txt = page.begin_text(final_x, final_y)
        x, y, alignment = 0, self.height - style.leading, style.text_align
        txt.set_position(x, y)
        txt.set_font(style.font_name, style.font_size, style.leading)
        for line in self.children:
            if self.style.text_align == TextAlign.right:
                txt.set_position(self.width - line.width)
            elif self.style.text_align == TextAlign.center:
                txt.set_position((self.width - line.width) / 2.0)
            fragments = []
            for word in line.words:
                if word is Break:
                    continue
                for part in word.parts:
                    if style != part.style:
                        if fragments:
                            txt.draw(''.join(fragments))
                            fragments = []
                        style = part.style
                        txt.set_font(style.font_name, style.font_size, style.leading)
                    fragments.append(part.part)
                fragments.append(' ')
            if fragments and fragments[-1] == ' ':
                fragments.pop()  # last space
            txt.draw(''.join(fragments), new_line=True)
        txt.close()
        return final_x, final_y



from .block import Block
from .text import Paragraph
from .style import Style, BorderCollapse


__all__ = ['Table', 'ColumnGroup', 'Column', 'RowGroup']


class ColumnGroup:

    def __init__(self, tag, parent, id=None, classes=None, css=None, position=None):
        self.tag = tag
        self.parent = parent
        self.id = id
        self.classes = classes or []
        self.css = css
        self.position = position
        self.style = Style.default()
        self.children = []
        self.widths = None
        self.span = None
        parent.columns = self

    def add(self, child):
        self.children.append(child)

    def get_widths(self, page, available_width):
        if self.widths is None:
            if self.span is None:
                self.calculate_widths(page, available_width)
            else:
                self.widths = [available_width/self.span] * self.span
        return self.widths

    def calculate_widths(self, page, available_width):

        widths = []

        for c in self.children:
            width = 0
            if c.is_fixed:
                width = c.fixed
            elif c.is_max_content:
                width = c.text_width(page, available_width)
            elif c.is_min_content:
                width = c.text_width(page, 0)
            widths.append(width)

        fixed = sum(widths)

        units = 0
        for c in self.children:
            if c.is_proportional:
                units += c.proportion

        unit_size = (available_width - fixed) / units
        for i in range(len(self.children)):
            c = self.children[i]
            if c.is_proportional:
                widths[i] = c.proportion * unit_size

        self.widths = widths


class Column(Paragraph):

    def __init__(self, *args):
        super().__init__(*args)
        self.width_spec = ''

    @property
    def is_fixed(self):
        return all(n in '0123456789' for n in self.width_spec)

    @property
    def fixed(self):
        return int(self.width_spec)

    @property
    def is_max_content(self):
        return self.width_spec in ('0*', 'max-content')

    @property
    def is_min_content(self):
        return self.width_spec in ('min-content',)

    @property
    def is_proportional(self):
        return self.width_spec.endswith('*') and self.width_spec != '0*'

    @property
    def proportion(self):
        return int(self.width_spec[:-1])

    def text_width(self, page, available_width):
        if self.words:
            return self.wrap(page, available_width)[0]
        else:
            return 0


class RowGroup:

    def __init__(self, tag, parent, id=None, classes=None, css=None, position=None):
        self.tag = tag
        self.parent = parent
        self.id = id
        self.classes = classes or []
        self.css = css
        self.position = 1
        self.style = Style.default()
        self.children = []
        self.drawn_on = None
        setattr(parent, tag, self)

    def add(self, child):
        self.children.append(child)

    def wrap(self, page, available_width):
        header_height = 0
        self.css.apply(self)
        if self.parent:
            self.style = self.style.inherit(self.parent.style)
        for row in self.children:
            header_height += row._wrap(page, available_width)[1]
        return header_height

    def draw(self, page, x, y):
        self.drawn_on = page.page_number
        for row in self.children:
            x, y = row._draw(page, x, y)
        self.position += 1
        return x, y


class Table(Block):

    __slots__ = ('columns', 'thead', 'tbody', 'tfoot', 'buffered', 'row_heights')

    def __init__(self, *args):
        super().__init__(*args)
        self.columns = None
        self.thead = None
        self.tbody = None
        self.tfoot = None
        self.buffered = False
        self.css.apply(self)
        if self.parent:
            self.style = self.style.inherit(self.parent.style)

    def add(self, child):
        if self.buffered:
            super().add(child)

    @classmethod
    def make(cls, parent=None, id=None, classes=None, style=None):
        return cls('table', parent, id, classes, style)

    def get_tbody(self):
        if not self.tbody:
            self.tbody = RowGroup('tbody', self, css=self.css)
        return self.tbody

    def get_column_widths(self, page, row, available_width):
        return self.columns.get_widths(page, available_width)

    def reserve_header_height(self, page, available_width):
        if self.thead and self.thead.drawn_on != page.page_number:
            return self.thead.wrap(page, available_width)
        return 0

    def reserve_footer_height(self, page, available_width):
        if self.tfoot:
            return self.tfoot.wrap(page, available_width)
        return 0

    def wrap(self, page, available_width):
        self.css.apply(self)
        if self.parent:
            self.style = self.style.inherit(self.parent.style)
        self.height = (
            self.reserve_header_height(page, available_width) +
            self.tbody.wrap(page, available_width) +
            self.reserve_header_height(page, available_width)
        )
        self.width = available_width
        return self.width, self.height

    def draw_header(self, page, x, y):
        if self.thead and self.thead.drawn_on != page.page_number:
            return self.thead.draw(page, x, y)
        return x, y

    def draw_footer(self, page, x, y):
        if self.tfoot:
            return self.tfoot.draw(page, x, y)
        return x, y

    def draw(self, page, x, y):
        x, y = self.draw_header(page, x, y)
        x, y = self.tbody.draw(page, x, y)
        x, y = self.draw_footer(page, x, y)
        return x, y

    def old_wrap(self, available_width, _=None):
        collapsed = self.style.border_collapse == BorderCollapse.collapse
        columns_available_width = available_width - self.frame_width
        vspacing = hspacing = 0
        if not collapsed:
            vspacing = self.style.border_spacing_vertical
            hspacing = self.style.border_spacing_horizontal
            columns_available_width -= hspacing*len(self.columns)
        self.calculate_ratio_columns(columns_available_width)
        self.width = available_width
        self.height = self.frame_height
        self.content_heights = []
        for row in self.rows:
            row.collapsed = collapsed
            row.cell_spacing = hspacing
            _, height = row.wrap(self.content_widths)
            height += vspacing
            self.content_heights.append(height)
            self.height += height
        return available_width, self.height

    def old_split(self, available_width, available_height):
        if self.content_heights is None:
            self.wrap(available_width)

        assert self.content_heights, "Split called on empty table."

        if self.tfoot:

            # check that footer fits, exit fast if not
            footer_height = sum(self.content_heights[-len(self.tfoot):])
            if footer_height >= available_height:
                return []

            # reduce the available height by footer to make sure there is always room for it
            available_height -= footer_height

        heights = iter(self.content_heights)

        # now check if header also fits, exit fast if not
        header_height = 0
        for _, height in zip(self.thead, heights):
            header_height += height
            if header_height >= available_height:
                return []

        # finally lets see if some rows fit sandwiched between thead/tfoot
        split_at_row = -1
        split_row_height = 0
        consumed_height = header_height
        for split_at_row, (row, split_row_height) in enumerate(zip(self.content, heights)):
            consumed_height += split_row_height
            if consumed_height >= available_height:
                break

        if consumed_height <= available_height:
            if len(self.content) == 1 or len(self.content)-1 == split_at_row:
                # nothing to split
                return [self]
            else:
                # table split cleanly between rows, no need to further split any rows/cells
                return [
                    Table(self.ratios, self.columns, self.content[:split_at_row+1], self.style, self.thead, self.tfoot, was_split=True),
                    Table(self.ratios, self.columns, self.content[split_at_row-1:], self.style, self.thead, self.tfoot, was_split=True)
                ]

        top_rows = self.content[:split_at_row]
        row = self.content[split_at_row]
        bottom_rows = self.content[split_at_row+1:]

        top_height = available_height - (consumed_height - split_row_height)
        if self.style.border_collapse == BorderCollapse.separate:
            top_height -= self.style.border_spacing_vertical
        split = row.splitOn(None, available_width, top_height)

        if not split:
            # cell could not be split,
            # move entirely to bottom half
            bottom_rows.insert(0, row)
        elif len(split) == 1:
            # cell completely fits in top half,
            # add placeholder to bottom half
            top_rows.append(row)
        elif len(split) == 2:
            top_rows.append(split[0])
            bottom_rows.insert(0, split[1])
        else:
            raise LayoutError("Splitting row {} produced unexpected result.".format(split_at_row))

        return [] if not top_rows else [
            Table(self.ratios, self.columns, top_rows, self.style, self.thead, self.tfoot, was_split=True),
            Table(self.ratios, self.columns, bottom_rows, self.style, self.thead, self.tfoot, was_split=True)
        ]



from pdfrw.objects import PdfObject, PdfString


def serialize(meta):
    if isinstance(meta, dict):
        yield b'<<'
        for key, value in meta.items():
            if not key.startswith('/'):
                key = '/'+key
            yield '{} '.format(key).encode()
            yield from serialize(value)
            yield b' '
        yield b'>>'
    elif isinstance(meta, (PdfObject, PdfString)):
        yield str(meta).encode()
    elif isinstance(meta, str):
        if meta.startswith('(') and meta.endswith(')'):
            pass
        elif not meta.startswith('/'):
            meta = '/'+meta
        yield meta.encode()
    elif isinstance(meta, list):
        yield b'[ '
        for value in meta:
            yield from serialize(value)
            yield b' '
        yield b']'
    else:
        yield str(meta).encode()


class PDFReference:

    def __init__(self, id, meta=None, name=None):
        self.id = id
        self.meta = meta or {}
        self.gen = 0
        self.chunks = []
        self.offset = None
        self.name = name

    def update(self, dict):
        self.meta.update(dict)

    def write(self, chunk):
        self.meta.setdefault('Length', 0)
        self.meta['Length'] += len(chunk)
        self.chunks.append(chunk)

    def read(self):
        yield "{r.id} {r.gen} obj\n".format(r=self).encode()
        yield from self.read_meta()
        if self.chunks and not self.chunks[-1].endswith(b'\n'):
            self.write(b'\n')
        if self.chunks:
            yield b"stream\n"
            yield from self.chunks
            yield b"endstream\n"
            del self.chunks
        yield b"endobj\n"

    def read_meta(self):
        yield from serialize(self.meta)
        yield b"\n"

    @property
    def has_content(self):
        return bool(self.chunks)

    def __str__(self):
        return "{r.id} {r.gen} R".format(r=self)



from .document import PDFDocument


class PDFStreamer:

    def __init__(self, block_generator, letterhead=None, layout='portrait'):
        self.generator = block_generator
        self.pdf = PDFDocument(block_generator.css, letterhead, layout)

    def __iter__(self):
        yield from self.pdf.header()
        page = None
        for block in self.generator:
            while block:
                page_created = False
                if page is None:
                    page = self.pdf.add_page()
                    page_created = True
                _, requested_height = block.wrap(page, page.available_width)
                if block.style.page_break_before:
                    if not page_created:
                        yield from page.read()
                        page = self.pdf.add_page()
                        _, requested_height = block.wrap(page, page.available_width)
                if requested_height <= page.available_height:
                    page.x, page.y = block.draw(page, page.x, page.y)
                    if block.style.page_break_after:
                        yield from page.read()
                        page = None
                    break
                else:
                    remainder, block = block.split(None, None, page.available_height)
                    if remainder:
                        _, requested_height = remainder.wrap(page, page.available_width)
                        #assert requested_height <= page.available_height
                        page.x, page.y = remainder.draw(page, page.x, page.y)
                    assert page.has_content
                    yield from page.read()
                    page = None
        if page:
            yield from page.read()
        yield from self.pdf.footer()



from reportlab.pdfbase.pdfmetrics import getFont, Font
from reportlab.pdfbase.ttfonts import TTFont, makeToUnicodeCMap

__all__ = ('getFont', 'read_font')


def read_font(doc, font):
    if isinstance(font, TTFont):
        yield from read_truetype(doc, font)
    elif isinstance(font, Font):
        yield from read_type1(doc, font)
    else:
        raise RuntimeError


def read_type1(doc, font):
    name = doc.fontMapping[font.fontName]
    ref = doc.font_references[name]
    ref.update({
        'Type': 'Font',
        'Subtype': 'Type1',
        'BaseFont': font.fontName,
        'Encoding': 'WinAnsiEncoding'
    })
    yield from doc.finalize_reference(ref)


# PDF font flags (see PDF Reference Guide table 5.19)
FF_SYMBOLIC = 1 << 3-1
FF_NONSYMBOLIC = 1 << 6-1


def SUBSETN(n, table=bytes.maketrans(b'0123456789', b'ABCDEFGIJK')):
    return bytes('%6.6d' % n, 'ASCII').translate(table)


def read_truetype(doc, font):
    state = font.state[doc]
    for n, subset in enumerate(state.subsets):
        internalName = font.getSubsetInternalName(n, doc)[1:]
        baseFontName = (b''.join((SUBSETN(n), b'+', font.face.name, font.face.subfontNameX))).decode('pdfdoc')

        fontFile = doc.ref()
        fontFile.write(font.face.makeSubset(subset))
        fontFile.meta['Length1'] = fontFile.meta['Length']
        yield from doc.finalize_reference(fontFile)

        flags = font.face.flags & ~ FF_NONSYMBOLIC
        flags = flags | FF_SYMBOLIC

        FontDescriptor = doc.ref({
            'Type': '/FontDescriptor',
            'Ascent': font.face.ascent,
            'CapHeight': font.face.capHeight,
            'Descent': font.face.descent,
            'Flags': flags,
            'FontBBox': font.face.bbox,
            'FontName': baseFontName,
            'ItalicAngle': font.face.italicAngle,
            'StemV': font.face.stemV,
            'FontFile2': fontFile,
        })
        yield from doc.finalize_reference(FontDescriptor)

        cmapStream = doc.ref()
        cmapStream.write(makeToUnicodeCMap(baseFontName, subset).encode())
        yield from doc.finalize_reference(cmapStream)

        font_ref = doc.font_references['/'+internalName]
        font_ref.update({
            'Type': 'Font',
            'Subtype': 'TrueType',
            'Name': internalName,
            'BaseFont': baseFontName,
            'FirstChar': 0,
            'LastChar': len(subset) - 1,
            'Widths': list(map(font.face.getCharWidth, subset)),
            'ToUnicode': cmapStream,
            'FontDescriptor': FontDescriptor,
        })
        yield from doc.finalize_reference(font_ref)



from bericht.node import Block
from .style import VerticalAlign
from .cell import Cell

__all__ = ('Row',)


class Row(Block):

    __slots__ = (
        'collapsed', 'cell_widths', 'cell_spacing',
    )

    def __init__(self, *args):
        super().__init__(*args)
        self.collapsed = False
        self.cell_widths = []
        self.cell_spacing = 0

    @property
    def table(self):
        return self.parent.parent

    @property
    def frame_top(self):
        if self.collapsed:
            return self.style.border_top_width
        return 0

    @property
    def frame_right(self):
        return 0

    @property
    def frame_bottom(self):
        if self.collapsed:
            return self.style.border_bottom_width
        return 0

    @property
    def frame_left(self):
        return 0

    def wrap(self, page, available_width):
        return available_width, (
            self.table.reserve_header_height(page, available_width) +
            self._wrap(page, available_width)[1] +
            self.table.reserve_footer_height(page, available_width)
        )

    def _wrap(self, page, available_width):
        column_widths = self.table.get_column_widths(page, self, available_width)
        self.width = available_width
        cell_widths = self.cell_widths = []

        self.css.apply(self)
        if self.parent:
            self.style = self.style.inherit(self.parent.style)

        if not self.children:
            return 0, 0

        col_width = iter(column_widths)
        for cell in self.children:
            if not cell or cell.colspan == 1:
                cell_widths.append(next(col_width))
            else:
                cell_widths.append(0)
                for col in range(cell.colspan):
                    cell_widths[-1] += next(col_width) + self.cell_spacing / 2.0

        max_height = 0
        for cell, cell_width in zip(self.children, cell_widths):
            if cell is not None:
                cell.collapsed = self.collapsed
                _, height = cell.wrap(page, cell_width)
                max_height = max(max_height, height)

        for cell in self.children:
            if not cell: continue
            cell.height = max_height

        self.height = self.frame_height + max_height
        return self.width, self.height

    def split(self, top_parent, bottom_parent, available_height):
        if available_height >= self.height:
            return self, None

        content_height = available_height - self.frame_height
        top_half = Row(self.tag, self.parent, self.id, self.classes, self.css, self.position)
        bottom_half = Row(self.tag, self.parent, self.id, self.classes, self.css, self.position+1)

        for cell in self.children:
            if cell:
                cell.split(top_half, bottom_half, content_height)

        return top_half, bottom_half

    def draw(self, page, x, y):
        x, y = self.table.draw_header(page, x, y)
        x, y = self._draw(page, x, y)
        return x, y

    def _draw(self, page, x, y):
        original_x, original_y = x, y
        x += self.cell_spacing / 2.0
        #y -= self.height
        y -= self.frame_top
        last = len(self.cell_widths)-1
        for i, (cell, width) in enumerate(zip(self.children, self.cell_widths)):
            if cell:
                if False and self.collapsed:
                    self.draw_collapsed_cell_border(
                        None if i == 0 else self.children[i-1],
                        cell,
                        None if i == last else self.children[i+1],
                    )
                else:
                    cell.draw_border_and_background(page, x, original_y - self.height)
                cell.draw(page, x, y)
            x += width + self.cell_spacing / 2.0
        return original_x, original_y - self.height

    def draw_collapsed_cell_border(self, before, cell, after):
        pass




from .reference import PDFReference
from .letterhead import PDFLetterhead
from .page import PDFPage
from .font import read_font

__all__ = ('PDFDocument',)


class PDFDocument:

    def __init__(self, css, letterhead=None, layout='portrait'):
        self.css = css
        self.ref_ids = 0
        self.offset = 0
        self.references = []
        self.page = None
        self.layout = layout
        self.root = self.ref({
            'Type': 'Catalog',
            'Pages': self.ref({
                'Type': 'Pages',
                'Count': 0,
                'Kids': []
            })
        })
        self.info = self.ref({
            'Producer': '(bericht)',
        })
        self.letterhead = PDFLetterhead(self, letterhead) if letterhead else None

        # font id (/F1) -> Reference object (7 0 R)
        self.font_references = {}
        # required by reportlab fonts handling code:
        self.fontMapping = {}  # font name (Helvetica) -> font id (/F1)
        self.delayedFonts = []  # fonts to be rendered

    def header(self):
        src = b"%PDF-1.7\n%\xc3\xbf\xc3\xbf\xc3\xbf\xc3\xbf\n"
        self.offset += len(src)
        yield src
        if self.letterhead:
            yield from self.letterhead.read()

    def finalize_reference(self, ref):
        ref.offset = self.offset
        for chunk in ref.read():
            self.offset += len(chunk)
            yield chunk

    @property
    def footer_refs(self):
        yield self.info
        yield self.root
        yield self.root.meta['Pages']

    def footer(self):
        for ref in self.footer_refs:
            yield from self.finalize_reference(ref)

        for font in self.delayedFonts:
            yield from read_font(self, font)

        yield b"xref\n"
        yield "0 {}\n".format(len(self.references) + 1).encode()
        yield b"0000000000 65535 f \n"
        for ref in self.references:
            yield "{:0>10} 00000 n \n".format(ref.offset).encode()

        yield b"trailer\n"
        yield from self.ref({
            'Size': len(self.references) + 1,
            'Root': self.root,
            'Info': self.info
        }).read_meta()

        yield b"startxref\n"
        yield str(self.offset).encode()
        yield b"\n%%EOF\n"

    def ref(self, meta=None, name=None):
        self.ref_ids += 1
        ref = PDFReference(self.ref_ids, meta, name)
        self.references.append(ref)
        return ref

    def add_page(self):
        pages = self.root.meta['Pages'].meta
        pages['Count'] += 1
        self.page = PDFPage(self, pages['Count'], self.css, 'a4', self.layout)
        pages['Kids'].append(self.page.dictionary)
        return self.page



from tinycss2 import (
    parse_stylesheet, parse_declaration_list, parse_one_component_value,
    ast, nth, color3
)
from bericht.node.style import Style

__all__ = ('CSS',)


class Combinator:

    DESCENDANT = 1          # E F    an F element descendant of an E element
    CHILD = 2               # E > F  an F element child of an E element
    NEXT_SIBLING = 3        # E + F  an F element immediately preceded by an E element
    SUBSEQUENT_SIBLING = 4  # E ~ F  an F element preceded by an E element

    __slots__ = ('type', 'simple_selectors')

    def __init__(self):
        self.type = None
        self.simple_selectors = []

    @staticmethod
    def select_tag(tag):
        return lambda node: tag == node.tag

    @staticmethod
    def select_id(node_id):
        return lambda node: node_id == node.id

    @staticmethod
    def select_class(name):
        return lambda node: name in node.classes

    @staticmethod
    def select_position(a_b):
        a, b = a_b

        def position_test(node):
            if a == 0:  # nth-child(0n+b)
                return node.position == b
            else:  # nth-child(an+b)
                return node.position >= b and (node.position-b) % a == 0

        return position_test

    @staticmethod
    def select_last_child(last_child):
        return lambda node: last_child == node.last_child

    def add(self, selector, value, negate=False):
        matcher = selector(value)
        if negate:
            self.simple_selectors.append(lambda node: not matcher(node))
        else:
            self.simple_selectors.append(matcher)

    def matches(self, node):
        for simple_selector in self.simple_selectors:
            if not simple_selector(node):
                return False
        return True


class Selector:

    __slots__ = ('combinators',)

    def __init__(self, combinators):
        self.combinators = combinators

    def matches(self, node):
        for combinator in reversed(self.combinators):
            if not combinator.matches(node):
                return False
            node = node.parent
        return True


class Declarations:

    def __init__(self, content):
        self.content = content
        self.attrs = {}
        for declaration in parse_declaration_list(content, skip_comments=True, skip_whitespace=True):
            if isinstance(declaration, ast.AtRule):
                if declaration.at_keyword == 'bottom-right':
                    at_dec = Declarations(declaration.content)
                    self.attrs['page_bottom_right_content'] = at_dec.attrs['content']
                else:
                    raise NotImplementedError
            else:
                value_component = parse_one_component_value(declaration.value, skip_comments=True)
                if 'color' in declaration.name:
                    self.attrs[declaration.name.replace('-', '_')] = color3.parse_color(value_component)
                elif 'content' == declaration.name:
                    self.attrs[declaration.name.replace('-', '_')] = parse_content_value(declaration.value)
                else:
                    self.attrs[declaration.name.replace('-', '_')] = value_component.value

    def apply(self, node):
        if node.style:
            node.style = node.style.set(**self.attrs)
        else:
            parent = Style.default() if node.parent is None else node.parent.style
            node.style = parent.set(**self.attrs)


def parse_content_value(value):
    parts = []
    for part in value:
        if isinstance(part, ast.StringToken):
            parts.append(part.value)
        elif isinstance(part, ast.FunctionBlock):
            if part.lower_name == 'counter' and part.arguments[0].lower_value == 'page':
                parts.append(lambda page: page.page_number)
            else:
                raise NotImplementedError
        elif isinstance(part, ast.WhitespaceToken):
            pass
        else:
            raise NotImplementedError

    return lambda page: ''.join(
        [p if isinstance(p, str) else str(p(page)) for p in parts]
    )


def parse_selectors(prelude):
    selectors = []
    combinator = None
    combinators = []
    prefix = None

    for token in prelude:
        if isinstance(token, ast.WhitespaceToken):
            combinator = None
        elif token == ',':
            selectors.append(Selector(combinators))
            combinator = None
            combinators = []
        elif token == '>':
            combinators[-1].type = Combinator.CHILD
            combinator = None
        elif token == '+':
            combinators[-1].type = Combinator.NEXT_SIBLING
            combinator = None
        elif token == '~':
            combinators[-1].type = Combinator.SUBSEQUENT_SIBLING
            combinator = None
        else:
            if not combinator:
                combinator = Combinator()
                combinators.append(combinator)
            if token in ('.', '#', ':'):
                if not prefix:
                    prefix = token.value
                else:
                    prefix += token.value
            elif isinstance(token, ast.IdentToken):
                if prefix is None:
                    combinator.add(combinator.select_tag, token.value)
                elif prefix == '.':
                    combinator.add(combinator.select_class, token.value)
                elif prefix == '#':
                    combinator.add(combinator.select_id, token.value)
                elif prefix == ':':
                    if token.value == 'first':
                        combinator.add(combinator.select_position, (0, 1))
                    elif token.value in ('last', 'last-child'):
                        combinator.add(combinator.select_last_child, True)
                    else:
                        raise NotImplementedError
                else:
                    raise NotImplementedError
                prefix = None
            elif isinstance(token, ast.FunctionBlock):
                prefix = None
                if token.name == 'nth-child':
                    combinator.add(combinator.select_position, nth.parse_nth(token.arguments))
                elif token.name == 'nth-of-type':
                    raise NotImplementedError
            else:
                raise NotImplementedError

    selectors.append(Selector(combinators))

    return selectors


class CSS:

    def __init__(self, src):
        self.src = src
        self.rules = []
        for rule in parse_stylesheet(src, skip_comments=True, skip_whitespace=True):
            if isinstance(rule, ast.AtRule) and rule.at_keyword == 'page':
                selectors = parse_selectors(rule.prelude)
                for selector in selectors:
                    for combinator in selector.combinators:
                        combinator.add(combinator.select_tag, '@page')
                    if not selector.combinators:
                        combinator = Combinator()
                        combinator.add(combinator.select_tag, '@page')
                        selector.combinators.append(combinator)
                self.rules.append(
                    (selectors, Declarations(rule.content))
                )
            elif isinstance(rule, ast.QualifiedRule):
                self.rules.append(
                    (parse_selectors(rule.prelude), Declarations(rule.content))
                )

    def apply(self, node):
        for selectors, declarations in self.rules:
            for selector in selectors:
                if selector.matches(node):
                    declarations.apply(node)



from .style import Style

__all__ = ('Block',)


class Block:

    __slots__ = (
        'tag', 'parent', 'id', 'classes', 'style', 'children',
        'css', 'position', 'last_child', 'width', 'height',
    )

    def __init__(self, tag, parent, id, classes, css, position, children=None):
        self.tag = tag
        self.parent = parent
        self.id = id
        self.classes = classes or []
        self.css = css
        self.position = position
        self.children = children or []
        self.width = None
        self.height = None
        self.last_child = False
        self.style = Style.default()
        if self.parent:
            self.parent.add(self)

    def add(self, child):
        self.children.append(child)

    def wrap(self, page, available_width):
        raise NotImplementedError

    def split(self, available_height):
        raise NotImplementedError

    def draw(self, page, x, y):
        raise NotImplementedError

    def draw_border_and_background(self, page, x, y):
        page.save_state()
        page.translate(x, y)
        s = self.style
        top_left, top_right, bottom_left, bottom_right = self.border_box
        if s.background_color:
            page.fill_color(*s.background_color)
            page.rectangle(0, 0, self.width, self.height)
        if s.border_top_width > 0:
            page.line_width(s.border_top_width)
            page.stroke_color(*s.border_top_color)
            page.line(*top_left, *top_right)
        if s.border_right_width > 0:
            page.line_width(s.border_right_width)
            page.stroke_color(*s.border_right_color)
            page.line(*top_right, *bottom_right)
        if s.border_bottom_width > 0:
            page.line_width(s.border_bottom_width)
            page.stroke_color(*s.border_bottom_color)
            page.line(*bottom_left, *bottom_right)
        if s.border_left_width > 0:
            page.line_width(s.border_left_width)
            page.stroke_color(*s.border_left_color)
            page.line(*top_left, *bottom_left)
        page.restore_state()

    @property
    def empty(self):
        return not self.children

    @property
    def min_content_width(self):
        width = 0
        for block in self.children:
            width = max(width, block.min_content_width)
        return width

    @property
    def frame_top(self):
        s = self.style
        return s.margin_top + s.border_top_width + s.padding_top

    @property
    def frame_right(self):
        s = self.style
        return s.padding_right + s.border_right_width + s.margin_right

    @property
    def frame_bottom(self):
        s = self.style
        return s.margin_bottom + s.border_bottom_width + s.padding_bottom

    @property
    def frame_left(self):
        s = self.style
        return s.margin_left + s.border_left_width + s.padding_left

    @property
    def frame_width(self):
        return self.frame_left + self.frame_right

    @property
    def frame_height(self):
        return self.frame_top + self.frame_bottom

    @property
    def border_box(self):
        s = self.style
        return (
            (s.margin_left, self.height-s.margin_top),  # top left
            (self.width-s.margin_right, self.height-s.margin_top),  # top right
            (s.margin_left, s.margin_bottom),  # bottom left
            (self.width-s.margin_right, s.margin_bottom),  # bottom right
        )



from .font import getFont
from reportlab.lib.rl_accel import escapePDF


class PDFText:

    def __init__(self, page, x=0, y=0):
        self.page = page
        self.write("BT\n")
        self.set_position(x, y)
        self.font_name = None
        self.font_size = None
        self.font_leading = None
        self.font_subset = None

    def write(self, code):
        self.page.write(code)

    def set_position(self, x=0, y=0):
        self.write("{} {} Td\n".format(x, y))

    def set_font(self, font_name, size, leading):
        self.font_name = font_name
        self.font_size = size
        self.font_leading = leading
        self.font_subset = None

    def draw(self, txt, new_line=False):
        doc = self.page.document
        font = getFont(self.font_name)
        if font._dynamicFont:
            for subset, t in font.splitString(txt, doc):
                if self.font_subset != subset:
                    name = font.getSubsetInternalName(subset, doc)
                    if name not in doc.font_references:
                        doc.font_references[name] = doc.ref()
                    if name not in self.page.font:
                        self.page.font[name] = doc.font_references[name]
                    self.write("{} {} Tf {} TL\n".format(
                        name,
                        self.font_size, self.font_leading))
                self.write("(%s) Tj " % escapePDF(t))
        elif font._multiByte:
            name = doc.fontMapping.get(font.fontName)
            if name is None:
                name = doc.fontMapping[font.fontName] = '/F{}'.format(len(doc.fontMapping)+1)
                doc.delayedFonts.append(font)
            if name not in doc.font_references:
                doc.font_references[name] = doc.ref()
            if name not in self.page.font:
                self.page.font[name] = doc.font_references[name]
            self.write("/{} {} Tf {} TL\n".format(name, self.font_size, self.font_leading))
            self.write("(%s) Tj " % font.formatForPdf(txt))
        else:
            raise RuntimeError

        if new_line:
            self.write('T*\n')

    def close(self):
        self.write("ET\n")



from .text import PDFText
from bericht.node.text import stringWidth

__all__ = ('PDFPage',)


class PDFPage:

    tag = '@page'

    def __init__(self, document, page_number, css, size='letter', layout='portrait'):
        self.parent = None
        self.classes = []
        self.style = None
        self.position = page_number
        css.apply(self)

        self.document = document
        self.size = size
        self.layout = layout

        dimensions = SIZES[self.size.upper()]
        self.width = dimensions[0 if self.layout is 'portrait' else 1]
        self.height = dimensions[1 if self.layout is 'portrait' else 0]

        margins = DEFAULT_MARGINS.copy()
        if self.style:
            for side in ('top', 'right', 'bottom', 'left'):
                attr = 'margin_'+side
                if attr in self.style.set_attrs:
                    margins[side] = getattr(self.style, attr)
        self.x = margins['left']
        self.y = self.height - margins['top']
        self.available_width = self.width - self.x - margins['right']
        self.margins = margins

        self.content = document.ref()

        # Initialize the Font, XObject, and ExtGState dictionaries
        self.resources = document.ref({
            'ProcSet': ['PDF', 'Text', 'ImageB', 'ImageC', 'ImageI'],
        })

        if self.style and 'letterhead_page' in self.style.set_attrs:
            letterhead_page = self.document.letterhead[int(self.style.letterhead_page)-1]
            self.resources.meta.update({
                'XObject': {letterhead_page.name: letterhead_page}
            })
            self.write('/{} Do\n'.format(letterhead_page.name))

        # The page dictionary
        self.dictionary = document.ref({
            'Type': 'Page',
            'Parent': document.root.meta['Pages'],
            'MediaBox': [0, 0, self.width, self.height],
            'Contents': self.content,
            'Resources': self.resources,
        })

        if self.style.page_bottom_right_content:
            style = self.style.set(font_size=9)
            page_text = self.style.page_bottom_right_content(self)
            text_width = stringWidth(page_text, style.font_name, style.font_size)
            text = self.begin_text(
                (self.margins['left'] + self.available_width) - text_width,
                self.margins['bottom'] - style.leading*2
            )
            text.set_font(style.font_name, style.leading, style.font_size)
            text.draw(page_text)
            text.close()

    @property
    def page_number(self):
        return self.position

    @property
    def has_content(self):
        return self.content.has_content

    @property
    def available_height(self):
        return self.y - self.margins['bottom']

    @property
    def font(self):
        if 'Font' not in self.resources.meta:
            self.resources.meta['Font'] = {}
        return self.resources.meta['Font']

    def write(self, chunk):
        self.content.write(chunk.encode())

    def read(self):
        for ref in (self.dictionary, self.resources, self.content):
            yield from self.document.finalize_reference(ref)

    def save_state(self):
        self.write("q\n")

    def restore_state(self):
        self.write("Q\n")

    def translate(self, x, y):
        self.write("1 0 0 1 {} {} cm\n".format(x, y))

    def begin_text(self, x, y):
        return PDFText(self, x, y)

    def line_width(self, width):
        self.write("{} w\n".format(width))

    def stroke_color(self, r, g, b, a):
        assert a == 1, "TODO: implement alpha"
        self.write("{} {} {} RG\n".format(r, g, b))

    def line(self, x1, y1, x2, y2):
        self.write("n {} {} m {} {} l S\n".format(x1, y1, x2, y2))

    def fill_color(self, r, g, b, a):
        assert a == 1, "TODO: implement alpha"
        self.write("{} {} {} rg\n".format(r, g, b))

    def rectangle(self, x, y, width, height):
        self.write("n {} {} {} {} re f*\n".format(x, y, width, height))


DEFAULT_MARGINS = {
    'top': 72,
    'left': 72,
    'bottom': 72,
    'right': 72
}

SIZES = {
    '4A0': [4767.87, 6740.79],
    '2A0': [3370.39, 4767.87],
    'A0': [2383.94, 3370.39],
    'A1': [1683.78, 2383.94],
    'A2': [1190.55, 1683.78],
    'A3': [841.89, 1190.55],
    'A4': [595.28, 841.89],
    'A5': [419.53, 595.28],
    'A6': [297.64, 419.53],
    'A7': [209.76, 297.64],
    'A8': [147.40, 209.76],
    'A9': [104.88, 147.40],
    'A10': [73.70, 104.88],
    'B0': [2834.65, 4008.19],
    'B1': [2004.09, 2834.65],
    'B2': [1417.32, 2004.09],
    'B3': [1000.63, 1417.32],
    'B4': [708.66, 1000.63],
    'B5': [498.90, 708.66],
    'B6': [354.33, 498.90],
    'B7': [249.45, 354.33],
    'B8': [175.75, 249.45],
    'B9': [124.72, 175.75],
    'B10': [87.87, 124.72],
    'C0': [2599.37, 3676.54],
    'C1': [1836.85, 2599.37],
    'C2': [1298.27, 1836.85],
    'C3': [918.43, 1298.27],
    'C4': [649.13, 918.43],
    'C5': [459.21, 649.13],
    'C6': [323.15, 459.21],
    'C7': [229.61, 323.15],
    'C8': [161.57, 229.61],
    'C9': [113.39, 161.57],
    'C10': [79.37, 113.39],
    'RA0': [2437.80, 3458.27],
    'RA1': [1729.13, 2437.80],
    'RA2': [1218.90, 1729.13],
    'RA3': [864.57, 1218.90],
    'RA4': [609.45, 864.57],
    'SRA0': [2551.18, 3628.35],
    'SRA1': [1814.17, 2551.18],
    'SRA2': [1275.59, 1814.17],
    'SRA3': [907.09, 1275.59],
    'SRA4': [637.80, 907.09],
    'EXECUTIVE': [521.86, 756.00],
    'FOLIO': [612.00, 936.00],
    'LEGAL': [612.00, 1008.00],
    'LETTER': [612.00, 792.00],
    'TABLOID': [792.00, 1224.00]
}


#!/usr/bin/env python
import os
from setuptools import setup, find_packages


BASE = os.path.dirname(__file__)
README_PATH = os.path.join(BASE, 'README.rst')
CHANGES_PATH = os.path.join(BASE, 'CHANGES.rst')
long_description = '\n\n'.join((
    open(README_PATH).read(),
    open(CHANGES_PATH).read(),
))


setup(
    name='bericht',
    version='0.0.6',
    url='https://github.com/systori/bericht',
    license='BSD',
    description='Improved tabular report generation with ReportLab.',
    long_description=long_description,
    author='Lex Berezhny',
    author_email='lex@damoti.com',
    keywords='pdf,table,report,reportlab,html',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: User Interfaces',
    ],
    install_requires=['reportlab', 'pdfrw', 'tinycss2'],
    packages=find_packages(exclude=('bericht.tests',)),
    include_package_data=True
)



from itertools import zip_longest
from bericht.table import *
from ..text import *

__all__ = ('TableBuilder',)


class TableBuilder:

    def __init__(self, ratios, style):
        self.ratios = ratios
        self.columns = [0]*len(ratios)
        self.thead = []
        self.tbody = []
        self.tfoot = []
        self.style = style

    def header(self, *row, cell_style=None, row_style=None):
        self._row(self.thead, *row, cell_style=cell_style, row_style=row_style)

    def row(self, *row, cell_style=None, row_style=None):
        self._row(self.tbody, *row, cell_style=cell_style, row_style=row_style)

    def footer(self, *row, cell_style=None, row_style=None):
        self._row(self.tfoot, *row, cell_style=cell_style, row_style=row_style)

    def _row(self, section, *row, cell_style, row_style):
        assert len(row) <= len(self.columns), "Too many cells added to row."
        cell_style = cell_style or self.style
        row_style = row_style or self.style
        cells = []
        for col, _ in zip_longest(row, self.columns):
            if isinstance(col, (Cell, Span)) or col is None:
                cell = col
            elif isinstance(col, (list, tuple)):
                cell = Cell(col, cell_style)
            elif isinstance(col, (Paragraph, Text, Table)):
                cell = Cell((col,), cell_style)
            else:
                cell = Cell((Text(str(col), cell_style),), cell_style)
            cells.append(cell)
        for i, (ratio, width, cell) in enumerate(zip(self.ratios, self.columns, cells)):
            if isinstance(cells[i], Cell) and ratio == 0:
                self.columns[i] = max(
                    width, cell.min_content_width+cell.frame_width
                )
        section.append(Row(cells, row_style))

    @property
    def table(self):
        return Table(self.ratios, self.columns, self.tbody, self.style, self.thead, self.tfoot)



from enum import Enum
from bericht.node import Block
from .style import VerticalAlign

__all__ = ('Cell',)


class Cell(Block):

    __slots__ = ('collapsed', 'first', 'last', 'content_heights', 'colspan')

    def __init__(self, *args, colspan=1):
        super().__init__(*args)
        self.collapsed = False
        self.content_heights = None
        self.colspan = colspan

    @property
    def frame_top(self):
        s = self.style
        if self.collapsed:
            return s.border_top_width/2 + s.padding_top
        else:
            return s.border_top_width + s.padding_top

    @property
    def frame_right(self):
        s = self.style
        if self.collapsed:
            return s.padding_right + s.border_right_width/2.0
        else:
            return s.padding_right + s.border_right_width

    @property
    def frame_bottom(self):
        s = self.style
        if self.collapsed:
            return s.border_bottom_width/2.0 + s.padding_bottom
        else:
            return s.border_bottom_width + s.padding_bottom

    @property
    def frame_left(self):
        s = self.style
        if self.collapsed:
            return s.border_left_width/2.0 + s.padding_left
        else:
            return s.border_left_width + s.padding_left

    @property
    def border_box(self):
        return (
            (0, self.height),  # top left
            (self.width, self.height),  # top right
            (0, 0),  # bottom left
            (self.width, 0),  # bottom right
        )

    def draw(self, page, x, y):
        x += self.frame_left
        if self.style.vertical_align == VerticalAlign.top:
            y -= self.frame_top
        elif self.style.vertical_align == VerticalAlign.middle:
            y -= (self.height + sum(self.content_heights)) / 2.0
        else:
            y -= self.height - (sum(self.content_heights) + self.frame_bottom)

        for block, height in zip(self.children, self.content_heights):
            block.draw(page, x, y)

    def wrap(self, page, available_width):
        self.width = available_width
        self.content_heights = []

        self.css.apply(self)
        if self.parent:
            self.style = self.style.inherit(self.parent.style)

        content_width = available_width - self.frame_width

        consumed = 0
        for block in self.children:
            _, height = block.wrap(page, content_width)
            consumed += height
            self.content_heights.append(height)

        self.height = self.frame_height + consumed
        return available_width, self.height

    def split(self, top_parent, bottom_parent, available_height):

        if not self.content_heights:
            self.parent = top_parent
            top_parent.children.append(self)
            bottom_parent.children.append(None)
            return

        if self.style.vertical_align == VerticalAlign.top:
            consumed_height = self.frame_top
        elif self.style.vertical_align == VerticalAlign.middle:
            consumed_height = (self.height - sum(self.content_heights)) / 2.0
        else:
            consumed_height = self.height - (sum(self.content_heights) + self.frame_bottom)

        if consumed_height >= available_height:
            # border and padding don't even fit
            top_parent.children.append(None)
            self.parent = bottom_parent
            bottom_parent.children.append(self)
            return

        split_at_block, split_block_height = -1, 0
        for split_at_block, split_block_height in enumerate(self.content_heights):
            consumed_height += split_block_height
            if consumed_height >= available_height:
                break

        if consumed_height <= available_height:
            if len(self.children) == 1 or len(self.children)-1 == split_at_block:
                self.parent = top_parent
                top_parent.children.append(self)
                bottom_parent.children.append(None)
                return
            else:
                top_cell = Cell(self.tag, top_parent, self.id, self.classes, self.css, self.position, colspan=self.colspan)
                for top_child in self.children[:split_at_block+1]:
                    top_child.parent = top_cell
                    top_cell.children.append(top_child)
                bottom_cell = Cell(self.tag, bottom_parent, self.id, self.classes, self.css, self.position, colspan=self.colspan)
                for bottom_child in self.children[split_at_block-1:]:
                    bottom_child.parent = bottom_cell
                    bottom_cell.children.append(bottom_child)
                return top_cell, bottom_cell

        top_cell = Cell(self.tag, top_parent, self.id, self.classes, self.css, self.position, colspan=self.colspan)
        for top_child in self.children[:split_at_block]:
            top_child.parent = top_cell
            top_cell.children.append(top_child)

        block = self.children[split_at_block]
        bottom_cell = Cell(self.tag, bottom_parent, self.id, self.classes, self.css, self.position, colspan=self.colspan)

        block.split(top_cell, bottom_cell, available_height - (consumed_height - split_block_height))

        for bottom_child in self.children[split_at_block+1:]:
            bottom_child.parent = bottom_cell
            bottom_cell.children.append(bottom_child)

        return top_cell, bottom_cell



from enum import Enum
from collections import namedtuple

from reportlab.lib.fonts import tt2ps
from reportlab.lib.colors import Color, black


__all__ = (
    'Style', 'TextAlign', 'VerticalAlign',
    'BorderStyle', 'BorderCollapse',
)


class TextAlign:
    left = 'left'
    center = 'center'
    right = 'right'
    justify = 'justify'


class VerticalAlign:
    top = 'top'
    middle = 'middle'
    bottom = 'bottom'


class BorderStyle(Enum):
    solid = 1
    dashed = 2


class BorderCollapse(Enum):
    collapse = 1
    separate = 2


_style_cache = {}


class Style(namedtuple('_Style', (

        'set_attrs',

        'font', 'font_size', 'font_weight', 'font_style', 'text_decoration',

        'text_align', 'vertical_align',

        'padding_top', 'padding_right', 'padding_bottom', 'padding_left',

        'border_top_color', 'border_top_style', 'border_top_width',
        'border_left_color', 'border_left_style', 'border_left_width',
        'border_right_color', 'border_right_style', 'border_right_width',
        'border_bottom_color', 'border_bottom_style', 'border_bottom_width',

        'margin_top', 'margin_right', 'margin_bottom', 'margin_left',

        'background_color',

        'page_break_before',
        'page_break_inside',
        'page_break_after',

        'page_bottom_right_content',

        'border_collapse',
        'border_spacing_horizontal',
        'border_spacing_vertical',

        'letterhead_page'

        ))):

    @classmethod
    def default(cls):
        return cls(

            set_attrs=tuple(),

            font='OpenSans',
            font_size=12,
            font_weight='',
            font_style='',
            text_decoration='',

            text_align='left',

            vertical_align='top',

            padding_top=0,
            padding_right=0,
            padding_bottom=0,
            padding_left=0,

            border_top_color=(0, 0, 0, 1),
            border_top_style=BorderStyle.solid,
            border_top_width=0,
            border_right_color=(0, 0, 0, 1),
            border_right_style=BorderStyle.solid,
            border_right_width=0,
            border_bottom_color=(0, 0, 0, 1),
            border_bottom_style=BorderStyle.solid,
            border_bottom_width=0,
            border_left_color=(0, 0, 0, 1),
            border_left_style=BorderStyle.solid,
            border_left_width=0,

            margin_top=0,
            margin_right=0,
            margin_bottom=0,
            margin_left=0,

            background_color=None,

            page_break_before=False,
            page_break_inside=True,
            page_break_after=False,

            page_bottom_right_content='',

            border_collapse=BorderCollapse.separate,
            border_spacing_horizontal=2,
            border_spacing_vertical=2,

            letterhead_page=1
        )

    def set(self, **attrs):
        attrs = self.convert(attrs)
        _ts = self._replace(
            set_attrs=tuple(set(self.set_attrs) | set(attrs.keys())),
            **attrs,
        )
        ts = _style_cache.setdefault(hash(_ts), _ts)
        if ts is _ts:  # validate if we just added this new style to cache
            ts.validate(**attrs)
        return ts

    def inherit(self, other):
        return self.set(**{
            attr: getattr(other, attr) for attr in other.set_attrs if attr not in self.set_attrs
        })

    @staticmethod
    def convert(kwargs):
        if 'page_break_before' in kwargs:
            kwargs['page_break_before'] = kwargs['page_break_before'] == 'always'
        if 'page_break_after' in kwargs:
            kwargs['page_break_after'] = kwargs['page_break_after'] == 'always'
        return kwargs

    def validate(self, **kwargs):
        for key, value in kwargs.items():
            if key in ('font', 'font_weight', 'font_style', 'text_decoration', 'text_align', 'vertical_align'):
                assert isinstance(value, str)
            elif key in ('font_size', 'border_spacing'):
                assert isinstance(value, (int, float))
            elif key in ('page_break_inside',):
                assert isinstance(value, bool)
            elif key == 'border_collapse':
                assert isinstance(value, BorderCollapse)
            elif key.startswith('border'):
                _, _, field = key.split('_')
                if field == 'color':
                    assert isinstance(value, Color)
                elif field == 'style':
                    assert isinstance(value, BorderStyle)
                elif field == 'width':
                    assert isinstance(value, (int, float))
            elif key.startswith('padding') or key.startswith('margin'):
                assert isinstance(value, (int, float))

    @property
    def bold(self):
        return 1 if self.font_weight == 'bold' else 0

    @property
    def italic(self):
        return 1 if self.font_style == 'italic' else 0

    @property
    def font_name(self):
        return tt2ps(self.font, self.bold, self.italic)

    @property
    def leading(self):
        return int(self.font_size * 1.2)



from pdfrw import PdfReader
from pdfrw.buildxobj import pagexobj

__all__ = ('PDFLetterhead',)


class PDFLetterhead:

    def __init__(self, document, path):
        self.document = document
        self.pages = []
        self.font_objects = []
        template = PdfReader(path)
        letterhead_num = 1
        for page in template.pages:
            template_page = pagexobj(page)
            fonts = self.get_fonts(template_page)
            form = document.ref({
                'Type': 'XObject',
                'Subtype': 'Form',
                'FormType': 1,
                'BBox': template_page.BBox,
                'Resources': {'Font': fonts}
            }, name='Letterhead{}'.format(letterhead_num))
            if template_page.Filter:
                form.meta['Filter'] = 'FlateDecode'
            if template_page.Resources.ExtGState:
                form.meta['Resources']['ExtGState'] = template_page.Resources.ExtGState
            form.write(template_page.stream.encode('latin'))
            self.pages.append(form)
            letterhead_num += 1

    def get_fonts(self, page):
        doc = self.document
        fonts = {}
        for font_id, font_ref in page.Resources.Font.iteritems():

            font = fonts[font_id] = doc.ref({
                key: font_ref[key] for key in
                set(font_ref.keys()) - {'/ToUnicode', '/FontDescriptor', '/DescendantFonts'}
            })
            self.font_objects.append(fonts[font_id])

            if font_ref.ToUnicode:
                font.meta['ToUnicode'] = doc.ref()
                font.meta['ToUnicode'].write(font_ref.ToUnicode.stream.encode('latin'))
                self.font_objects.append(font.meta['ToUnicode'])

            if font_ref.FontDescriptor:
                font.meta['FontDescriptor'] = self.get_font_descriptor(font_ref.FontDescriptor)

            if font_ref.DescendantFonts:
                dfs = font.meta['DescendantFonts'] = []
                for descendant in font_ref.DescendantFonts:
                    dfont = doc.ref({
                        key: descendant[key] for key in
                        set(descendant.keys()) - {'/FontDescriptor'}
                    })
                    dfont.meta['FontDescriptor'] = self.get_font_descriptor(descendant.FontDescriptor)
                    dfs.append(dfont)
                    self.font_objects.append(dfont)

        return fonts

    def get_font_descriptor(self, descriptor):
        fd = self.document.ref()
        for key, value in descriptor.iteritems():
            if key.startswith('/FontFile'):
                ff = self.document.ref({
                    key: value for key, value in value.iteritems()
                    if key in ('/Filter', '/Subtype')
                })
                ff.write(value.stream.encode('latin'))
                self.font_objects.append(ff)
                fd.meta[key] = ff
            elif key.startswith('/CIDSet'):
                cidset = self.document.ref({
                    key: value for key, value in value.iteritems()
                    if key in ('/Filter', '/Subtype')
                })
                cidset.write(value.stream.encode('latin'))
                self.font_objects.append(cidset)
                fd.meta[key] = cidset
            else:
                fd.meta[key] = value
        self.font_objects.append(fd)
        return fd

    def read(self):
        for obj in self.font_objects:
            yield from self.document.finalize_reference(obj)
        for page in self.pages:
            yield from self.document.finalize_reference(page)

    def __getitem__(self, item):
        if item < 1:
            return self.pages[0]
        if item >= len(self.pages):
            return self.pages[-1]
        return self.pages[item]



from string import whitespace
from lxml import etree
from bericht.node import Style
from bericht.node import Paragraph, Word, Break
from bericht.node import Table, Row, Cell
from bericht.node import ColumnGroup, Column, RowGroup

__all__ = ('HTMLParser',)


def pumper(source, pump, stream):
    while True:
        for chunk in stream:
            yield chunk
        try:
            pump(next(source))
        except StopIteration:
            break


TAG_MAP = {
    'p': Paragraph,
    'table': Table,
    'colgroup': ColumnGroup,
    'bericht-col': Column,
    'thead': RowGroup,
    'tbody': RowGroup,
    'tfoot': RowGroup,
    'tr': Row,
    'td': Cell,
}

BLOCK_TAGS = tuple(TAG_MAP.keys())
STYLE_TAGS = 'b', 'strong', 'i', 'u', 'span', 'br'


class HTMLParser:

    def __init__(self, html_generator, css=None):
        parser = etree.HTMLPullParser(
            events=('start', 'end',),
            tag=BLOCK_TAGS + STYLE_TAGS,
            remove_comments=True
        )
        self.events = pumper(
            html_generator,
            parser.feed,
            parser.read_events()
        )
        self.css = css

    def traverse(self, parent=None):
        for event, element in self.events:
            tag = element.tag

            assert tag in BLOCK_TAGS

            if event == 'start':

                if tag == 'tr' and isinstance(parent, Table):
                    parent = parent.get_tbody()

                node = TAG_MAP[tag](
                    tag, parent,
                    element.attrib.get('id', None),
                    element.attrib.get('class', None),
                    self.css,
                    1
                )

                if isinstance(node, Table):
                    if parent is None:
                        row_iter = self.traverse(node)
                        current_row = next(row_iter)
                        positions = {}
                        for next_row in row_iter:
                            positions.setdefault(current_row.parent.tag, -1)
                            positions[current_row.parent.tag] += 1
                            next_row.position = positions[current_row.parent.tag]
                            if current_row.parent.tag == 'tbody':
                                yield current_row
                            current_row = next_row
                        for cell in current_row.children:
                            if cell.children:
                                cell.children[-1].last_child = True
                        if current_row.children:
                            current_row.children[-1].last_child = True
                        current_row.last_child = True
                        yield current_row
                    else:
                        node.buffered = True
                        for _ in self.traverse(node):
                            pass
                        yield node

                elif isinstance(node, Paragraph):
                    self.fast_forward_to_end(element)
                    node.style = Style.default()
                    extract_words(node, element)
                    if isinstance(node, Column):
                        node.width_spec = element.attrib.get('width', '')
                    yield node

                else:

                    if isinstance(node, Cell):
                        node.colspan = int(element.attrib.get('colspan', 1))

                    elif isinstance(node, ColumnGroup):
                        span = element.attrib.get('span')
                        if span:
                            node.span = int(span)

                    for _ in self.traverse(node):
                        pass

                    yield node

            else:

                break

    def fast_forward_to_end(self, end):
        """ Fast forwarding (parsing) to end tag loads
            the element and its contents into memory.
            Sometimes it's necessary to have the entire
            element tree loaded to correctly process
            the contents (eg, <p></p> tags).
        """
        for event, element in self.events:
            if event == 'end' and element == end:
                return
        raise RuntimeError('End element not reached.')

    def __iter__(self):
        yield from self.traverse()


def extract_words(p, root):

    words = p.words
    styles = [p.style]
    word_open = False

    def data(data):
        nonlocal word_open
        if data[0] in whitespace or (words and words[-1] is Break):
            word_open = False
        parts = data.split()
        if not parts:
            return
        word_iter = iter(parts)
        style = styles[-1]
        if word_open:
            words[-1].add(p, style, next(word_iter))
        words.extend(Word(p, style, word) for word in word_iter)
        word_open = data[-1] not in whitespace

    if root.text:
        data(root.text)

    context = etree.iterwalk(
        root, events=('start', 'end'),
        tag=STYLE_TAGS
    )
    for event, element in context:
        tag = element.tag
        if event == 'start':
            if tag in ('b', 'strong'):
                styles.append(styles[-1].set(font_weight='bold'))
            elif tag == 'i':
                styles.append(styles[-1].set(font_style='italic'))
            elif tag == 'u':
                styles.append(styles[-1].set(text_decoration='underline'))
            if element.text:
                data(element.text)
        else:  # event == 'end':
            if tag == 'br':
                words.append(Break)
            else:
                styles.pop()
            if element.tail:
                data(element.tail)
